#lang scheme

;; the source machine, a tiny model of CML-style choice.
;; note that without replication, these source terms must always
;; converge.

;; this omission may or may not be significant; during Q&A at the ICFP talk,
;; Reppy suggested that this technique may not be compositional.

(require redex/reduction-semantics)

(provide sourcelang s-> denotes)


(define-language sourcelang
  (prog (s ...))
  (s a
     (select a ...))
  (a (in chan) (out chan))
  (chan variable))

(define s->
  (reduction-relation
   sourcelang
   (--> (s_1 ... (select a_1 ... (in chan) a_2 ...) s_2 ... (select a_3 ... (out chan) a_4 ...) s_3 ...)
        (s_1 ... (in chan) s_2 ... (out chan) s_3 ...))
   (--> (s_1 ... (select a_1 ... (out chan) a_2 ...) s_2 ... (select a_3 ... (in chan) a_4 ...) s_3 ...)
        (s_1 ... (out chan) s_2 ... (in chan) s_3 ...))))




;; modeling choices: model parallel composition using a list.


;; the denotation of a source term (essentially, what actions have been completed).
(define-metafunction sourcelang
  denotes : prog -> (a ...)
  [(denotes (s ...)) (a ... ...)
                     (where ((a ...) ...) ((denotes1 s) ...))])


(define-metafunction sourcelang
  denotes1 : s -> (a ...)
  [(denotes1 (select a ...)) ()]   ;; throw away selects
  [(denotes1 a) (a)]) ;; keep ins & outs


(test-equal
 (term (denotes ((select (in a) (out b)) (out c) (select (out d)) (in e))))
 (term ((out c) (in e))))

(let ([red-coverage (make-coverage s->)])
  (parameterize ([relation-coverage (list red-coverage)])
    (test-->
     s->
     (term ((select (in abc)) (select (out abc))))
     (term ((in abc) (out abc))))
    
    (test-->
     s->
     (term ((select (out abc)) (select (in abc))))
     (term ((out abc) (in abc))))
    
    (test--> 
     s->
     (term ((select (out def) (out abc) (out ghi)) (select (in jkl) (in abc) (in mno))))
     (term ((out abc) (in abc))))
    
    (test-->
     s->
     (term ((select (in jkl) (in abc) (in mno)) (select (out def) (out abc) (out ghi))))
     (term ((in abc) (out abc))))
    
    
    (test-->
     s->
     (term ((select (in jkl) (in abc) (out def)) (select (in def) (out abc) (out ghi))))
     (term ((out def) (in def)))
     (term ((in abc) (out abc))))
    
    ;; the example from the paper
    (test-->>
     s->
     (term ((select (in x) (in y)) (select (in z)) (select (out x)) (select (out y) (out z))))
     (term ((in y) (select (in z)) (select (out x)) (out y)))
     (term ((in x) (in z) (out x) (out z))))
    
    ;; uncomment to see coverage results.
    #;(covered-cases red-coverage)
    ))
