
#lang scheme

(require redex/reduction-semantics
         "target.rkt"
         "can-reach.rkt")


;; TESTS


;; tests of no-dups

(test-equal
 (term (no-dups (a b c d e e d c a)))
 (term (a b c d e)))

;; tests of overlap

(test-equal
 (term (overlap (a b x c d) (e f g x h i)))
 (term (x)))

(test-equal
 (term (overlap (a b x y c) (d e x y f)))
 (term (x)))

;; tests of free-vars
(test-equal
 (term (free-vars (active b (in ch1))))
 (term (b)))

(test-equal 
 (term (free-vars (nu (a b c) ())))
 (term ()))

;; tests of rename
(test-equal
 (term (rename b b1 ((active b (in ch1)) (active c (out ch1)))))
 (term ((active b1 (in ch1)) (active c (out ch1)))))

;; tests for clean

;; no change required:
(test-equal
 (term (clean (nu (a b c) ((active b (in ch1)) (active c (out ch1)))) 
              (active d (in ch1))))
 (term (nu (a b c) ((active b (in ch1)) (active c (out ch1))))))

;; collision:
(test-equal
 (term (clean (nu (a b c) ((active b (in ch1)) (active c (out ch1)))) 
              (active b (in ch1))))
 (term (nu (a b1 c) ((active b1 (in ch1)) (active c (out ch1))))))

(test-equal
 (term (clean (nu (a b c) ()) (nu (d e f) ())))
 (term (nu (a b c) ())))

(test-equal
 (term (clean (nu (d e f) ()) ()))
 (term (nu (d e f) ())))

;; another kind of collision;
(test-equal
 (term (clean
        (nu (a2) ((heart a2) (heart a1)))
        ((heart a2))))
 (term (nu (a3) ((heart a3) (heart a1)))))


;; tests for interleave:
(test-equal
 (term (interleave () ()))
 (term ()))

(test-equal
 (term (interleave ((active a (in ch1)) (in ch2) (selected ((a (out ch1)) (c (in ch4))) b))
                   ((active b1 (in ch2)) (active b1 (out ch1)) (closed ((c1 (in ch2)) (b1 (out ch1)))))))
 (term ((active a (in ch1)) (active b1 (in ch2)) (active b1 (out ch1)) (in ch2) (closed ((c1 (in ch2)) (b1 (out ch1)))) (selected ((a (out ch1)) (c (in ch4))) b))))

;; tests for nu-combine

(test-equal
 (term (nu-combine (nu (a b c) ()) (nu (d e f) ())))
 (term (nu (a b c d e f) ())))

(test-equal
 (term (nu-combine (nu (a b c) ((active a (in ch1)) (in ch2) (selected ((a (out ch1)) (c (in ch4))) b)))
                   (nu (a b c) ((active b (in ch2)) (active b (out ch1)) (closed ((c (in ch2)) (b (out ch1))))))))
 (term (nu (a b c a1 b1 c1)
           ((active a (in ch1)) (active b1 (in ch2)) (active b1 (out ch1)) (in ch2) (closed ((c1 (in ch2)) (b1 (out ch1)))) (selected ((a (out ch1)) (c (in ch4))) b)))))


#;(test-->
 t->
 (term ((active z (out c1)) (free c1) (active y (in c1))))
 (term (tsort ((heart z) (heart y) (announced c1 y z) (free c1)))))

;; tsort

(test-equal
 (term (tsort ()))
 (term (nu () ())))

(test-equal 
 (term (tsort (in ch1)))
 (term (nu () ((in ch1)))))


(test-equal
 (term (tsort ((in ch1) (active p1 (out ch2)) (nu (p1 p2) ((closed ((p1 (in ch2)) (p2 (out ch2)))))))))
 (term (nu (p3 p2) ((active p1 (out ch2)) (in ch1) (closed ((p3 (in ch2)) (p2 (out ch2))))))))


(time 
 (test-predicate
  (lambda (x) x)
  (can-reach
   t->
   (term (tsort (nu (a b c d) ((free ch1) 
                              (free ch2) 
                              (active a (in ch1))
                              (active b (in ch2))
                              (active c (out ch1))
                              (active d (out ch2))
                              (open ((a (in ch1)) (b (in ch2)))) (open ((c (out ch1)) (d (out ch2))))))))
   ;; let's say a and c succeed:
   (term (tsort (nu (a b c d) ((free ch1)
                              (free ch2)
                              (in ch1)
                              (out ch1)
                              (active b (in ch2))
                              (active d (out ch2))
                              (closed ((a (in ch1))
                                       (b (in ch2))))
                              (closed ((c (out ch1))
                                       (d (out ch2))))))))
   ;; and then both c and d can be refused:
   ;; oops, this one spins:
   (term (tsort (nu (a b c d)
                   ((free ch1)
                    (free ch2)
                    (announced ch2 b d)
                    (refused b)
                    (refused d)
                    (in ch1)
                    (out ch1)
                    (closed ((a (in ch1))
                             (b (in ch2))))
                    (closed ((c (out ch1))
                             (d (out ch2)))))))))))

;; the main reduction relation:

(define out-xy-sync (term ((xout (out x)) (yout (out y)))))
(define in-x-sync (term ((xin (in x)))))

  


(printf "about to start main tests.\n")
    
(let ([red-coverage (make-coverage t->)])
  (parameterize ([relation-coverage (list red-coverage)])
    
    (test-->
     t->
     (term (tsort (nu (a b c d) ((active c (out c1)) (free c1) (active b (in c1))))))
     (term (tsort (nu (a b c d) ((heart b) (heart c) (announced c1 b c) (free c1))))))
    
    (test-->
     t->
     (term (tsort (nu (a b c d) ((heart b) (heart c) (announced c1 b c) (free c1))))))
    
    (test-->
     t->
     (term (tsort (nu (a b c d) ((heart b) (heart c) (announced c1 b c) (free c1) (open ((a (in c1)) (b (in c1)) (c (out c1))))))))
     (term (tsort (nu (a b c d) ((selected ((a (in c1)) (b (in c1)) (c (out c1))) b) (heart c) (announced c1 b c) (free c1) (closed ((a (in c1)) (b (in c1)) (c (out c1))))))))
     (term (tsort (nu (a b c d) ((selected ((a (in c1)) (b (in c1)) (c (out c1))) c) (heart b) (announced c1 b c) (free c1) (closed ((a (in c1)) (b (in c1)) (c (out c1)))))))))
    
    
    
    ;; unsuccessful communication (both part of the same synchronizer)
    (test-predicate
     (lambda (x) x)
     (can-reach
      t->
      (term (tsort (nu (a b c d) ((active a (in c1)) (active b (in c1)) (active c (out c1)) (free c1) (open ((a (in c1)) (b (in c1)) (c (out c1))))))))
      ;; a and c get chosen:
      (term (tsort (nu (a b c d) ((heart a) (active b (in c1)) (heart c) (announced c1 a c) (free c1) (open ((a (in c1)) (b (in c1)) (c (out c1))))))))
      ;; or maybe b and c get chosen:
      (term (tsort (nu (a b c d) ((active a (in c1)) (heart b) (heart c) (announced c1 b c) (free c1) (open ((a (in c1)) (b (in c1)) (c (out c1))))))))
      ;; now, either b or c can be blessed by the synchronizer. Let's say it's b:
      (term (tsort (nu (a b c d) ((active a (in c1)) (selected ((a (in c1)) (b (in c1)) (c (out c1))) b) (heart c) (announced c1 b c) (free c1) (closed ((a (in c1)) (b (in c1)) (c (out c1))))))))
      ;; unfortunately, this means that c gets refused:
      (term (tsort (nu (a b c d) ((active a (in c1)) (selected ((a (in c1)) (b (in c1)) (c (out c1))) b) (refused c) (announced c1 b c) (free c1) (closed ((a (in c1)) (b (in c1)) (c (out c1))))))))
      ;; now, the announced communication gets canceled.
      (term (tsort (nu (a b c d) ((active a (in c1)) (canceled ((a (in c1)) (b (in c1)) (c (out c1)))) (free c1) (closed ((a (in c1)) (b (in c1)) (c (out c1))))))))
      ;;; ... and everything starts all over again, with an extra floating "closed" synchronizer (oops, c1 and c1 ...)
      (term (tsort (nu (a b c d a1 b1 c1) ((active a (in c1)) 
                                           (active a1 (in c1))
                                           (active b1 (in c1))
                                           (active c1 (out c1))
                                           (free c1)
                                           (open ((a1 (in c1)) (b1 (in c1)) (c1 (out c1))))
                                           (closed ((a (in c1)) (b (in c1)) (c (out c1))))))))))
    
    (test-predicate
     (lambda (x) x)
     (can-reach
      t->
      (term (tsort (nu (a b c d) ((selected ((a (in c1)) (b (in c1)) (c (out c1))) b) (heart c) (announced c1 b c) (free c1) (closed ((a (in c1)) (b (in c1)) (c (out c1))))))))
      ;; c's synchronizer is closed:
      (term (tsort (nu (a b c d) ((selected ((a (in c1)) (b (in c1)) (c (out c1))) b) (refused c) (announced c1 b c) (free c1) (closed ((a (in c1)) (b (in c1)) (c (out c1))))))))
      ;; c is refused, go to canceled:
      (term (tsort (nu (a b c d) ((canceled ((a (in c1)) (b (in c1)) (c (out c1)))) (free c1) (closed ((a (in c1)) (b (in c1)) (c (out c1))))))))
      ))
    
    ;; both refused?
    (time
     (test-predicate
     (lambda (x) x)
     (can-reach
      t->
      (term (tsort (nu (a b c d) ((free ch1) 
                                  (free ch2) 
                                  (active a (in ch1))
                                  (active b (in ch2))
                                  (active c (out ch1))
                                  (active d (out ch2))
                                  (open ((a (in ch1)) (b (in ch2)))) (open ((c (out ch1)) (d (out ch2))))))))
      ;; let's say a and c succeed:
      (term (tsort (nu (a b c d) ((free ch1)
                                  (free ch2)
                                  (in ch1)
                                  (out ch1)
                                  (active b (in ch2))
                                  (active d (out ch2))
                                  (closed ((a (in ch1))
                                           (b (in ch2))))
                                  (closed ((c (out ch1))
                                           (d (out ch2))))))))
      ;; and then both c and d can be refused:
      (term (tsort (nu (a b c d) ((free ch1)
                                  (free ch2)
                                  (announced ch2 b d)
                                  (refused b)
                                  (refused d)
                                  (in ch1)
                                  (out ch1)
                                  (closed ((a (in ch1))
                                           (b (in ch2))))
                                  (closed ((c (out ch1))
                                           (d (out ch2)))))))))))
    
    
    ;; other order in selected:
    (test-->
     t->
     (term (tsort (nu (a b c d) ((selected ((a (in c1)) (b (in c1)) (c (out c1))) b) (refused c) (announced c1 c b) (free c1) (closed ((a (in c1)) (b (in c1)) (c (out c1))))))))
     ;; c is refused, go to canceled:
     (term (tsort (nu (a b c d) ((canceled ((a (in c1)) (b (in c1)) (c (out c1)))) (free c1) (closed ((a (in c1)) (b (in c1)) (c (out c1)))))))))
    
    (test-->
     t->
     (term (tsort (nu (a b c d) ((canceled ((a (in c1)) (b (in c1)) (c (out c1)))) (free c1) (closed ((a (in c1)) (b (in c1)) (c (out c1))))))))
     (term (tsort (nu (a b c d) ((nu (a b c) ((active a (in c1)) (active b (in c1)) (active c (out c1)) (open  ((a (in c1)) (b (in c1)) (c (out c1)))))) (free c1) (closed ((a (in c1)) (b (in c1)) (c (out c1)))))))))
    
    
    
    (test-->>
     t->
     (term (tsort (nu (xout yout xin) ((free x) (free y) (active xout (out x)) (active yout (out y)) (active xin (in x)) (open ,out-xy-sync) (open ,in-x-sync)))))
     ;; in & out can be reordered, so can the two closed synchronizers:
     (term (tsort (nu (xout yout xin) ((free x) (free y) (in x) (out x) (active yout (out y)) (closed ,out-xy-sync) (closed ,in-x-sync)))))
     (term (tsort (nu (xout yout xin) ((free x) (free y) (in x) (out x) (active yout (out y)) (closed ,in-x-sync) (closed ,out-xy-sync)))))
     (term (tsort (nu (xout yout xin) ((free x) (free y) (out x) (in x) (active yout (out y)) (closed ,out-xy-sync) (closed ,in-x-sync)))))
     (term (tsort (nu (xout yout xin) ((free x) (free y) (out x) (in x) (active yout (out y)) (closed ,in-x-sync) (closed ,out-xy-sync))))))
    
    ;; uncomment the following line to 
    ;; see how many of the cases of t-> 
    ;; are covered
    #;
    (covered-cases red-coverage)))


(test-equal 
 (term (denotes/t (nu (a b c) ((in ch1) (heart c) (open ((a (in ch3)) (b (out ch4)))) (out ch2)))))
 (term ((in ch1) (out ch2))))
