#lang scheme

(require redex/reduction-semantics
         "grammar.rkt")

(define (well-typed? e)
  (match (term (typeof () ,e))
    [`(,_ ε) #t]
    [_ #f]))

(define-metafunction SR
  typeof : Γ e -> τ?
  [(typeof Γ (shift e))
   (A (@cps B C))
   ; Fig. 2 disallows control effects in e, though they should
   ; be safe.
   (where ((⇒ ((⇒ (A ε) (B ε)) ε) (C ε)) ε) (typeof Γ e))]
  
  [(typeof Γ (reset e))
   (C ε)
   (where (A (@cps A C)) (typeof Γ e))]
  [(typeof Γ (reset e))
   (A ε)
   (where (A ε) (typeof Γ e))]
  
  [(typeof Γ (e_1 e_2))
   (B δ)
   (where ((⇒ (A ε) (B γ)) α) (typeof Γ e_1))
   (where (C β) (typeof Γ e_2))
   (side-condition (term (= A C)))
   (where δ (comp α β γ))]
  
  [(typeof ((x_0 τ_0) ...) (block (def f (x : (A ε)) : (B β) = e) r ...))
   (typeof ((f ((⇒ (A ε) τ) ε)) (x_0 τ_0) ...) (block r ...))
   (where τ (typeof ((x (A ε)) (f ((⇒ (A ε) (B β)) ε)) (x_0 τ_0) ...) e))
   (side-condition (term (= (B β) τ)))]
  
  [(typeof ((x_0 τ_0) ...) (block (val x : (A ε) = e) r ...))
   (B γ)
   (where (C α) (typeof ((x_0 τ_0) ...) e))
   (side-condition (term (= C A)))
   (where (B β) (typeof ((x (A ε)) (x_0 τ_0) ...) (block r ...)))
   (where γ (comp α β))]
  
  [(typeof Γ (block e r_0 r_1 ...))
   (B γ)
   (where (A α) (typeof Γ e))
   (where (B β) (typeof Γ (block r_0 r_1 ...)))
   (where γ (comp α β))]
  
  [(typeof Γ (block e))
   (typeof Γ e)]
  
  [(typeof ((x_0 τ_0) ... (x_i τ_i) (x_i+1 τ_i+1) ...) x_i)
   τ_i
   (side-condition (term (not-in x_i (x_0 ...))))]
  [(typeof Γ inc)
   ((⇒ (Int ε) (Int ε)) ε)]
  [(typeof Γ dbl)
   ((⇒ (Int ε) (Int ε)) ε)]
  [(typeof Γ len)
   ((⇒ (Str ε) (Int ε)) ε)]
  
  [(typeof Γ string) (Str ε)]
  [(typeof Γ number) (Int ε)]
  
  [(typeof Γ e) #f])

(define-metafunction SR
  comp : α ... -> α?
  [(comp) ε]
  [(comp ε α ...) (comp α ...)]
  [(comp (@cps B C) α ...)
   (@cps B C)
   (where ε (comp α ...))]
  ; Consider adding subtyping.
  [(comp (@cps B C) α ...)
   (@cps V C)
   (where (@cps V W) (comp α ...))
   (side-condition (term (= B W)))]
  [(comp α ...) #f])

(define-metafunction SR
  [(= any any) #t]
  [(= any_!_1 any_!_1) #f])

(define-metafunction SR
  [(not-in any (any_0 ...))
   ,(not (member (term any) (term (any_0 ...))))])

(provide typeof comp well-typed?)